#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/cygwin.h>
#include <dlfcn.h>
#include <X11/Xlib.h>
#include <locale.h>
#include <pthread.h>
#include <math.h>

#define timerspec_add(a, b, result)                                           \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec + (b)->tv_sec;                             \
    (result)->tv_nsec = (a)->tv_nsec + (b)->tv_nsec;                          \
    if ((result)->tv_nsec >= 1000000000)                                      \
      {                                                                       \
        ++(result)->tv_sec;                                                   \
        (result)->tv_nsec -= 1000000000;                                      \
      }                                                                       \
  } while (0)


/* from libiberty */
char ** dupargv (char **argv);
void freeargv (char **vector);
char **buildargv (const char *input);
/* void expandargv (int *argcp, char ***argvp); */


/* So this is ONE way... */
#if 0

/* because we're linking against libiberty (for buildargv etc), we are
   ALSO forced to get getopt from it and NOT from cygwin.  If we 
   include getopt.h (or unistd.h), then we'd use declspec(dllimport)
   variables optind, optopt, and optarg -- but thanks to static 
   linkage, we're using libiberty's getopt() and getopt_long.

   libiberty's functions use static versions of optind, optopt, and
   optarg, NOT the declspec ones.  So, to make sure that both we 
   and libiberty are talking about the same extern variables:

   (1) do NOT #include unistd.h or getopt.h  !!!
   (2) define all the stuff from getopt.h right here.
*/
extern int opterr;
extern int optind;
extern int optopt;
extern int optreset;
extern char *optarg;
int getopt (int, char * const *, const char *);
struct option {
        const char *name;
        int  has_arg;
        int *flag;
        int val;
};
int getopt_long (int, char *const *, const char *, const struct option *, int *);
int getopt_long_only (int, char *const *, const char *, const struct option *, int *);
#define no_argument             0
#define required_argument       1
#define optional_argument       2

#else
/* The other way is to go ahead and #include getopt.h, but force linking
   to the cygwin version of getopt:
     gcc -o foo.exe  foo.o -lcygwin -liberty
   Which turns into the linker command
     .... -o foo.exe  foo.o -lcygwin -liberty -lgcc -lcygwin ...
   with 2 -lcygwin.  It looks odd, but it's okay.
*/
#include <getopt.h>
#endif

#define WIN32_LEAD_AND_MEAN
#define NOMINMAX
#include <windows.h>

#define PATH_SEP_CHAR_STR "/"
#define SEP_CHAR ":"
#define SEP_CHARS ":"

#define XSERV_TIMEDOUT -1
#define XSERV_FOUND     0
#define XSERV_NOTFOUND  1

typedef Display*(*XOpenDisplayFP)(char*);
typedef int(*XCloseDisplayFP)(Display *);

typedef struct s_opts_ { 
  char* displayname;
  char* appendpath;
  char* prependpath;
  char* replacepath;
  char* xlibname;
} s_opts;

typedef struct WorkerThreadData_ {
  XOpenDisplayFP  xopendis;
  XCloseDisplayFP xclosedis;
  char*           displayname;
} WorkerThreadData;

int realMain(int argc, char* argv[]);
char* pfopen(char *retval, const char *name, const char *dirs);
int fileExists(char* fullname, const char* path, const char* name);
void message_(int level, char* fmt, ...);
void debug_(char* fmt, ...);
void dumpOpts(s_opts* s);
void cleanup(s_opts* s);
void computeTimespec(double val, timespec_t* t);

pthread_mutex_t mtx_xopenOK;
pthread_cond_t  cv_xopenOK;
int xopenOK = XSERV_TIMEDOUT;
int try_with_timeout(WorkerThreadData* data, double delay);
void* open_display(void* /* WorkerThreadData* */ v);

int opt_nogui = 0;
int opt_notty = 0;
int opt_debug = 0;
int opt_location = 0;
double opt_timeout = 0.5;
#if defined(SILENT)
int opt_silent = 1;
#else
int opt_silent = 0;
#endif

#if defined(DEBUG)
# define debugMsg(fmt, ...) debug_(fmt, ## __VA_ARGS__)
#else
# define debugMsg(fmt, ...) do { if (opt_debug) { debug_(fmt, ## __VA_ARGS__); } } while(0)
#endif

#define realMsg(fmt, ...)  message_(-1, fmt, ## __VA_ARGS__)

#define errorMsg(fmt, ...)  do { if (!opt_silent) { message_(2, fmt, ## __VA_ARGS__); } } while(0)
#define warnMsg(fmt, ...)   do { if (!opt_silent) { message_(1, fmt, ## __VA_ARGS__); } } while(0)
#define infoMsg(fmt, ...)   do { if (!opt_silent) { message_(0, fmt, ## __VA_ARGS__); } } while(0)

const char* XLIBfmt = "cygX11-%d.dll";
const char* DefaultAppendPath = "/usr/X11R6/bin" SEP_CHAR "/usr/bin";


#ifndef ENABLE_GUI
# define ENABLE_GUI 1
#endif
#ifndef ENABLE_TTY
# define ENABLE_TTY 1
#endif
#if (ENABLE_TTY == 0) && (ENABLE_GUI == 0)
# error At least one of ENABLE_GUI and ENABLE_TTY must be 1
#endif

void cleanup(s_opts* s)
{
  if (!s) return;
  if (s->displayname) { free(s->displayname); s->displayname = NULL; }
  if (s->appendpath)  { free(s->appendpath);  s->appendpath  = NULL; }
  if (s->prependpath) { free(s->prependpath); s->prependpath = NULL; }
  if (s->replacepath) { free(s->replacepath); s->replacepath = NULL; }
  if (s->xlibname)    { free(s->xlibname);    s->xlibname    = NULL; }
}
void dumpOpts(s_opts* s)
{
  if (s) {
    if (s->displayname) { debugMsg("displayname : '%s'", s->displayname); } 
    if (s->appendpath)  { debugMsg("appendpath  : '%s'", s->appendpath);  }
    if (s->prependpath) { debugMsg("prependpath : '%s'", s->prependpath); }
    if (s->replacepath) { debugMsg("replacepath : '%s'", s->replacepath); }
    if (s->xlibname)    { debugMsg("xlibname    : '%s'", s->xlibname);    }
  }
  debugMsg("opt_debug   : %d", opt_debug);
  debugMsg("opt_location: %d", opt_location);
  debugMsg("opt_silent  : %d", opt_silent);
  debugMsg("opt_nogui   : %d", opt_nogui);
  debugMsg("opt_notty   : %d", opt_notty);
  debugMsg("opt_timeout : %5.2f", opt_timeout);
}

int realMain(int argc, char* argv[])
{
  int i, c, count = 0;
  char* origPathList = NULL;
  char* pathList = NULL;
  char bufFullPath[MAX_PATH + FILENAME_MAX + 100];
  char bufFileName[FILENAME_MAX];
  void* handleToDll = NULL;
  XOpenDisplayFP xopendis = NULL;
  XCloseDisplayFP xclosedis = NULL;
  s_opts sopt;
  char* p;

  int rc = 0;
  sopt.displayname = NULL;
  sopt.appendpath = NULL;
  sopt.prependpath = NULL;
  sopt.replacepath = NULL;
  sopt.xlibname = NULL;

  /***********************
   * Option parsing
   ***********************/
  while (1) {
    static struct option long_options[] =
    {
      {"help", no_argument, NULL, 'h'},
      {"display", required_argument, NULL, 'd'},
      {"location", no_argument, NULL, 'l'},
      {"appendpath", required_argument, NULL, 'a'},
      {"prependpath", required_argument, NULL, 'p'},
      {"replacepath", required_argument, NULL, 'r'},
      {"xlibname", required_argument, NULL, 'x'},
      {"timeout", required_argument, NULL, 't'},
#if (ENABLE_GUI == 1)
      {"nogui", no_argument, &opt_nogui, 1},
#endif
#if (ENABLE_TTY == 1)
      {"notty", no_argument, &opt_notty, 1},
#endif
      {"debug", no_argument, &opt_debug, 1},
#if defined(SILENT)
      {"no-silent", no_argument, &opt_silent, 0},
#else
      {"silent", no_argument, &opt_silent, 1},
#endif
      {0, 0, 0, 0}
    };

    static char* helpString = 
    "checkX determines if X is installed and Xserver is running\n" 
    "  returns 0 if yes, nonzero otherwise\n\n"
    "Options:\n"
    "  -h|--help                 : prints this help message\n"
    "  -d|--display STRING       : use STRING instead of $DISPLAY\n"
    "  -l|--location             : print location of Xlib DLL on stdout\n"
    "  -a|--appendpath STRING    : append STRING to value of $PATH (cumulative)\n"
    "  -p|--prependpath STRING   : prepend STRING to value of $PATH (cumulative)\n"
    "  -r|--replacepath STRING   : use STRING instead of $PATH when searching\n"
    "  -x|--xlibname STRING      : use exactly STRING instead of fuzzy cygX11-n.dll search\n"
    "  -t|--timeout FLOAT        : allow FLOAT seconds to connect with Xserver\n" 
    "                              defaults to 0.5, use 0.0 for Xlib's (safe, 12s) timeout\n"
#if (ENABLE_GUI == 1)
    "     --nogui                : disable informational popups\n"
#endif
#if (ENABLE_TTY == 1)
    "     --notty                : disable stderr messages\n"
#endif
    "     --debug                : turn on debugging messages\n"
#if defined(SILENT)
    "     --no-silent            : allow error, warning, and info messages\n"
#else
    "     --silent               : suppress error, warning, and info messages\n"
#endif
    "Note that -a defaults to '/usr/bin:/usr/X11R6/bin'.  To eliminate the default,\n"
    "use '-a \"\"'\n";

    int option_index = 0;
    c = getopt_long(argc, argv, "hd:la:p:r:x:t:", long_options, &option_index);

    if (c == -1)
      break;

    switch (c) {
      case 0:
        /* If this option set a flag, do nothing else now. */
        if (long_options[option_index].flag != 0)
        break;
      case 'h':
        realMsg(helpString);
        cleanup(&sopt);
        return 1;
      case 'l':
        opt_location = 1;
        break;
      case 'd':
        if (sopt.displayname) { free(sopt.displayname); sopt.displayname = NULL; }
        sopt.displayname = strdup(optarg);
        if (!sopt.displayname) {
          errorMsg("out of memory");
          cleanup(&sopt);
          return 1;
        }
        break;
      case 'a':
        if (!sopt.appendpath) {
          sopt.appendpath = strdup(optarg);
          if (!sopt.appendpath) {
            errorMsg("out of memory");
            cleanup(&sopt);
            return 1;
          }
        }
        else {
          p = sopt.appendpath;
          sopt.appendpath = (char*) malloc(strlen(sopt.appendpath) + strlen(optarg) + 2);
          if (!sopt.appendpath) {
            errorMsg("out of memory");
            free(p);
            cleanup(&sopt);
            return 1;
          }
          strcpy(sopt.appendpath, p);
          strcat(sopt.appendpath, SEP_CHAR);
          strcat(sopt.appendpath, optarg);
        }
        break;
      case 'p':
        if (!sopt.prependpath) {
          sopt.prependpath = strdup(optarg);
          if (!sopt.prependpath) {
            errorMsg("out of memory");
            cleanup(&sopt);
            return 1;
          }
        }
        else {
          p = sopt.prependpath;
          sopt.prependpath = (char*) malloc(strlen(sopt.prependpath) + strlen(optarg) + 2);
          if (!sopt.prependpath) {
            errorMsg("out of memory");
            free(p);
            cleanup(&sopt);
            return 1;
          }
          strcpy(sopt.prependpath, optarg);
          strcat(sopt.prependpath, SEP_CHAR);
          strcat(sopt.prependpath, p);
        }
        break;
      case 'r':
        if (sopt.replacepath) { free(sopt.replacepath); sopt.replacepath = NULL; }
        sopt.replacepath = strdup(optarg);
        if (!sopt.replacepath) {
          errorMsg("out of memory");
          cleanup(&sopt);
          return 1;
        }
        break;
      case 'x':
        if (sopt.xlibname) { free(sopt.xlibname); sopt.xlibname = NULL; }
        sopt.xlibname = strdup(optarg);
        if (!sopt.xlibname) {
          errorMsg("out of memory");
          cleanup(&sopt);
          return 1;
        }
        break;
      case 't':
        opt_timeout = strtod(optarg, &p);
        if (optarg == p) {
          errorMsg("bad value specified for timeout: %s\n", optarg);
          cleanup(&sopt);
          return 1;
        }
        break;
      case '?':
        /* getopt_long already printed an error message. */
        cleanup(&sopt);
        return 1;
      default:
        errorMsg("unrecognized option");
        cleanup(&sopt);
        return 1;
    }
  }
  if ((opt_nogui || (!ENABLE_GUI)) && (opt_notty || (!ENABLE_TTY)))
    if (opt_debug)
      opt_notty = 0;
  
  dumpOpts(&sopt);


  /***********************
   * PATH handling
   ***********************/
  if (!sopt.appendpath) {
    sopt.appendpath = strdup(DefaultAppendPath);
    if (!sopt.appendpath) {
      errorMsg("out of memory");
      cleanup(&sopt);
      return 1;
    }
  }

  origPathList = getenv("PATH");
  c = strlen(origPathList);
  if (sopt.replacepath) { c  = strlen(sopt.replacepath);     }
  if (sopt.appendpath)  { c += strlen(sopt.appendpath) + 1;  }
  if (sopt.prependpath) { c += strlen(sopt.prependpath) + 1; }

  if ((pathList = (char*)malloc(c * sizeof(char*)) ) == NULL) {
     errorMsg("out of memory");
     cleanup(&sopt);
     return 1;
  }
  pathList[0] = '\0';
  if (sopt.prependpath) {
    strcat(pathList, sopt.prependpath);
    strcat(pathList, SEP_CHAR);
  }
  if (sopt.replacepath) {
    strcat(pathList, sopt.replacepath);
  } else {
    strcat(pathList, origPathList);
  }
  if (sopt.appendpath) {
    strcat(pathList, SEP_CHAR);
    strcat(pathList, sopt.appendpath);
  }
  debugMsg("path is : %s", pathList);


  /***********************
   * XLIB library name
   ***********************/
  if (sopt.xlibname) {
    pfopen(bufFullPath, sopt.xlibname, pathList);
  } else {
    for (i = 6; i < 10; i++) {
      sprintf(bufFileName, XLIBfmt, i);
      pfopen(bufFullPath, bufFileName, pathList);
      if (fileExists(NULL,NULL,bufFullPath))
         break;
      bufFullPath[0] = '\0'; 
    }
  }
  free( pathList );
  if (bufFullPath[0] == '\0') {
    errorMsg("could not locate Xlib DLL %s", (sopt.xlibname ? sopt.xlibname : XLIBfmt));
    cleanup(&sopt);
    return 1;
  }
  debugMsg("DLL is %s", bufFullPath);  


  /***********************
   * XLIB library and symbol load
   ***********************/
  handleToDll = dlopen(bufFullPath, RTLD_LAZY | RTLD_GLOBAL);
  if (!handleToDll) {
    errorMsg("problem loading %s: %s", bufFullPath, dlerror());
    cleanup(&sopt);
    return 1;
  } else {
    debugMsg("%s dlopen'ed successfully.", bufFullPath);
  }

  xopendis =  (XOpenDisplayFP) dlsym(handleToDll, "XOpenDisplay");
  if (!xopendis) {
    errorMsg("problem loading symbol XOpenDisplay: %s", dlerror());
    rc = 1;
  } else {
    debugMsg("symbol XOpenDisplay loaded ok");
 
    xclosedis = (XCloseDisplayFP)dlsym(handleToDll, "XCloseDisplay"); 
    if (!xclosedis) {
      errorMsg("problem loading symbol XCloseDisplay: %s", dlerror());
      rc = 1;
    } else {
      Display* dpy = NULL;
      char* disVar = NULL;

      debugMsg("symbol XCloseDisplay loaded ok");
      disVar = sopt.displayname ? sopt.displayname : getenv("DISPLAY");
      WorkerThreadData data;

      data.xopendis = xopendis;
      data.xclosedis = xclosedis;
      data.displayname = disVar;
 
      rc = try_with_timeout(&data, opt_timeout);

      if (rc == 0) {
        debugMsg("X display '%s' successfully opened",
                 (disVar ? disVar : ""));
      } else {
        errorMsg("could not open X display '%s'", 
                 (disVar ? disVar : ""));
      }
    }
  }
  dlclose(handleToDll);
  cleanup(&sopt);

  if (opt_location && (rc == 0)) {
    /* deliberately do NOT use message_() macros */
    fprintf(stdout, "%s", bufFullPath);
  }

  debugMsg("returning with status %d", rc);
  return rc; /* exits with status of 0 if X available */
}

void computeTimespec(double val, timespec_t* t)
{
   double frac;
   if (!t) return;
   if (val < 0) return;
   t->tv_sec = (time_t)(floor(val));
   frac = (val - (double)t->tv_sec);
   t->tv_nsec = (long)(1000000000.0 * frac);
}
int try_with_timeout(WorkerThreadData* data, double delay)
{
  pthread_attr_t attr;
  pthread_t      id;
  int            status;
  timespec_t     now;
  timespec_t     delta;
  timespec_t     then;
  
  computeTimespec(fabs(delay), &delta);
  debugMsg("Using delay of %d secs, %ld nanosecs (%5.2f)",  delta.tv_sec, delta.tv_nsec,
           (double)delta.tv_sec + ((double)delta.tv_nsec)/1000000000.0);

  pthread_mutex_init(&mtx_xopenOK, NULL);
  pthread_cond_init (&cv_xopenOK, NULL);
         
  pthread_attr_init(&attr);
  pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

  pthread_mutex_lock(&mtx_xopenOK);
 
  pthread_create(&id, &attr, open_display, (void*)data);
  pthread_attr_destroy(&attr);

  if (delay != 0.0) {
    clock_gettime(CLOCK_REALTIME, &now);
    timerspec_add(&now, &delta, &then);
    pthread_cond_timedwait (&cv_xopenOK, &mtx_xopenOK, &then);
  }

  pthread_mutex_unlock(&mtx_xopenOK);

  if (delay != 0.0) {
    pthread_detach(id);
  } else {
    pthread_join(id, (void*)&status);
  }

  pthread_mutex_destroy(&mtx_xopenOK);
  debugMsg("xserver search was %s", (xopenOK == XSERV_TIMEDOUT ? "timed out" :
                                     (xopenOK == XSERV_NOTFOUND ? "unsuccessful" : "a success!")));

  // translate timeout into not-found
  return (xopenOK == XSERV_TIMEDOUT ? XSERV_NOTFOUND : xopenOK );
}

void* open_display(void* /* WorkerThreadData* */ v)
{
  Display* dpy;
  int rc = 0;
  WorkerThreadData* data = (WorkerThreadData*)v;

  if( (dpy = (*(data->xopendis))(data->displayname)) == NULL ) {
    rc = 1;
  } else {
    (*(data->xclosedis))(dpy);
    rc = 0;
  }
  pthread_mutex_lock (&mtx_xopenOK);
  xopenOK = rc;
  pthread_cond_signal(&cv_xopenOK);
  pthread_mutex_unlock (&mtx_xopenOK);

  pthread_exit((void*)0);
}
/*
 * This function attempts to locate a file which may be in any of
 * several directories. Unlike the original pfopen, it does not
 * return a FILE pointer to the opened file, but rather returns
 * the fully-qualified filename of the first match found. Returns
 * empty string if not found.
 */
char *pfopen(char *retval, const char *name, const char *dirs)
{
    char *ptr;
    char *tdirs;
    char returnval[MAX_PATH + FILENAME_MAX + 100];
    char *recursive_name;
    int foundit = FALSE;
    
    returnval[0] = '\0';

    if (dirs == NULL || dirs[0] == '\0')
        return NULL;

    if ((tdirs = malloc(strlen(dirs)+1)) == NULL)
        return NULL;

    strcpy(tdirs, dirs);

    for (ptr = strtok(tdirs, SEP_CHARS); (foundit == FALSE) && ptr != NULL;
         ptr = strtok(NULL, SEP_CHARS))
    {
       foundit = fileExists(returnval,ptr,name);
    }

    free(tdirs);
    if (!foundit)
        retval[0] = '\0';
    else
        strcpy(retval,returnval);
    return retval;
}
int fileExists(char* fullname, const char* path, const char* name)
{
   int retval = FALSE;
   FILE* file;
   size_t len;
   char work[FILENAME_MAX];
   char work2[MAX_PATH + FILENAME_MAX + 100];
   size_t wlen = sizeof(work);
   if (path != NULL)
   {
      strncpy(work, path, wlen);
      work[sizeof(work)-1] = 0;
      len = strlen(work);
      wlen = sizeof(work) - len;
      if (len && work[len-1] != '/' && work[len-1] != '\\')
      {    
         strncat(work, PATH_SEP_CHAR_STR, wlen);
         work[sizeof(work)-1] = 0;
         wlen = sizeof(work) - strlen(work);
      }
   }
   else
      work[0]='\0';
   
   strncat(work, name, wlen);
   work[sizeof(work)-1] = 0;
   wlen = sizeof(work) - strlen(work);
   cygwin_conv_to_posix_path(work, work2); 

#ifdef DEBUGALL
   debugMsg("looking for...\t%s",work2);
#endif

   file = fopen(work2, "rb");
   if (file != NULL)
   {
      if (fullname != NULL)
         strcpy(fullname,work2);
      retval = TRUE;
      fclose(file);
   }
   return retval;
}
void message_(int level, char* fmt, ...)
{
   char buf[10000];
   int j;
   va_list args;
   va_start(args, fmt);
   j = vsprintf(buf,fmt,args);
   va_end(args);
   if (ENABLE_TTY && (!opt_notty)) {
     if (level == 2) 
       fprintf(stderr, "checkX Error: %s\n", buf);
     else if (level == 1)
       fprintf(stderr, "checkX Warning: %s\n", buf);
     else if (level == 0)
       fprintf(stderr, "checkX Info: %s\n", buf);
     else
       fprintf(stdout, buf);
     return;
   }
   if (ENABLE_GUI && (!opt_nogui)) {
     if (level == 2) 
       MessageBox(NULL, buf, "checkX Error", MB_ICONERROR);
     else if (level == 1)
       MessageBox(NULL, buf, "checkX Warning", MB_ICONWARNING);
     else if (level == 0)
       MessageBox(NULL, buf, "checkX Info", MB_ICONINFORMATION);
     else
       MessageBox(NULL, buf, "checkX", MB_ICONINFORMATION);
     return;
   }
}
void debug_(char* fmt, ...)
{
   char buf[10000];
   int j;
   va_list args;
   va_start(args, fmt);
   j = vsprintf(buf,fmt,args);
   va_end(args);

   if (ENABLE_TTY && (!opt_notty)) {
     fprintf(stderr, "checkX DEBUG: %s\n", buf);
     return;
   }
   if (ENABLE_GUI && (!opt_nogui)) {
     MessageBox(NULL, buf, "checkX DEBUG", MB_ICONINFORMATION);
     return;
   }
}
int countargv(char** vector)
{
  char **scan;
  int rc = 0;
  if (vector != NULL)
    for (scan = vector; *scan != NULL; scan++)
      rc++;
  return rc;
}

int WINAPI
WinMain (HINSTANCE hSelf, HINSTANCE hPrev, LPSTR cmdline, int nShow)
{
  char **argv, *p, **q;
  int argc;
  char bufArg0[MAX_PATH];
  char bufTemp[MAX_PATH];
  char* fullCmdLine;
  int rc, i;

  opt_debug = 0;
  opt_nogui = 0;
  opt_notty = 0;
#if defined(SILENT)
  opt_silent = 1;
#else
  opt_silent = 0;
#endif

  setlocale(LC_ALL, "C");

  debugMsg("orig cmdline = %s", cmdline);
  GetModuleFileName(NULL, bufTemp, sizeof(bufTemp));
  cygwin_conv_to_full_posix_path(bufTemp, bufArg0);

  fullCmdLine = (char*)malloc(strlen(bufArg0) + strlen(cmdline) + 4);  
  if (!fullCmdLine) {
    errorMsg("out of memory");
    return 1;
  }

  /* protect embedded spaces in bufArg0 */
  strcpy(fullCmdLine, "\"");
  strcat(fullCmdLine, bufArg0);
  strcat(fullCmdLine, "\" ");
  strcat(fullCmdLine, cmdline);

  debugMsg("new cmdline = %s", fullCmdLine);

  argv = buildargv(fullCmdLine);
  argc = countargv(argv);
  debugMsg("argc = %d", argc);
  if(argv) {
    q = argv;
    for (q = argv; *q != NULL; q++)
      debugMsg(": '%s'", *q);
  }

  for (i = 0; i < argc; i++) {
    debugMsg("argv[%d]: '%s'", i, argv[i]);
  }

  rc = realMain(argc, argv);
  freeargv(argv);

  infoMsg("Exiting with status %d", rc);
  // note that this will kill the detached thread if it's still running
  return rc;
}

